#   Code up to mate4dummies 0.4.2 (at time of branch) is Copyright © Stephan Zevenhuizen
#   MATE Objects, (10-05-2018).
#   MATE for Dummies, (20-08-2018).

"""
MATE for Dummies. MATE Scripting System in Python
"""

__version__ = '1.0'

__all__ = ['MATE','objects']

from .mate import *
